# -*- coding: utf-8 -*-
"""
Dependencies: flask, tornado

SeeAlso:
    routes.turk_identification
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from ibeis.control import accessor_decors, controller_inject
from ibeis.algo.hots import pipeline
from flask import url_for, request, current_app  # NOQA
from os.path import join, dirname, abspath, exists
import cv2
import numpy as np   # NOQA
import utool as ut
from ibeis.web import appfuncs as appf
from ibeis import constants as const
import traceback
import requests
import six
from datetime import datetime
ut.noinject('[apis_query]')


CLASS_INJECT_KEY, register_ibs_method = (
    controller_inject.make_ibs_register_decorator(__name__))
register_api   = controller_inject.get_ibeis_flask_api(__name__)
register_route = controller_inject.get_ibeis_flask_route(__name__)


GRAPH_CLIENT_PEEK = 100


ANNOT_INFR_PEAK_MAX = 50


@register_ibs_method
@accessor_decors.default_decorator
@register_api('/api/query/annot/rowid/', methods=['GET'])
def get_recognition_query_aids(ibs, is_known, species=None):
    """
    DEPCIRATE

    RESTful:
        Method: GET
        URL:    /api/query/annot/rowid/
    """
    qaid_list = ibs.get_valid_aids(is_known=is_known, species=species)
    return qaid_list


@register_ibs_method
@register_api('/api/query/chip/dict/simple/', methods=['GET'])
def query_chips_simple_dict(ibs, *args, **kwargs):
    r"""
    Runs query_chips, but returns a json compatible dictionary

    Args:
        same as query_chips

    RESTful:
        Method: GET
        URL:    /api/query/chip/dict/simple/

    SeeAlso:
        query_chips

    CommandLine:
        python -m ibeis.web.apis_query --test-query_chips_simple_dict:0
        python -m ibeis.web.apis_query --test-query_chips_simple_dict:1

        python -m ibeis.web.apis_query --test-query_chips_simple_dict:0 --humpbacks

    Example:
        >>> # xdoctest: +REQUIRES(--web)
        >>> from ibeis.control.IBEISControl import *  # NOQA
        >>> import ibeis
        >>> ibs = ibeis.opendb(defaultdb='testdb1')
        >>> #qaid = ibs.get_valid_aids()[0:3]
        >>> qaids = ibs.get_valid_aids()
        >>> daids = ibs.get_valid_aids()
        >>> dict_list = ibs.query_chips_simple_dict(qaids, daids)
        >>> qgids = ibs.get_annot_image_rowids(qaids)
        >>> qnids = ibs.get_annot_name_rowids(qaids)
        >>> for dict_, qgid, qnid in list(zip(dict_list, qgids, qnids)):
        >>>     dict_['qgid'] = qgid
        >>>     dict_['qnid'] = qnid
        >>>     dict_['dgid_list'] = ibs.get_annot_image_rowids(dict_['daid_list'])
        >>>     dict_['dnid_list'] = ibs.get_annot_name_rowids(dict_['daid_list'])
        >>>     dict_['dgname_list'] = ibs.get_image_gnames(dict_['dgid_list'])
        >>>     dict_['qgname'] = ibs.get_image_gnames(dict_['qgid'])
        >>> result  = ut.repr2(dict_list, nl=2, precision=2, hack_liststr=True)
        >>> result = result.replace('u\'', '"').replace('\'', '"')
        >>> print(result)

    Example:
        >>> # xdoctest: +REQUIRES(--web)
        >>> from ibeis.control.IBEISControl import *  # NOQA
        >>> import time
        >>> import ibeis
        >>> import requests
        >>> # Start up the web instance
        >>> web_instance = ibeis.opendb_in_background(db='testdb1', web=True, browser=False)
        >>> time.sleep(10)
        >>> web_port = ibs.get_web_port_via_scan()
        >>> if web_port is None:
        >>>     raise ValueError('IA web server is not running on any expected port')
        >>> baseurl = 'http://127.0.1.1:%s' % (web_port, )
        >>> data = dict(qaid_list=[1])
        >>> resp = requests.get(baseurl + '/api/query/chip/simple/dict/', data=data)
        >>> print(resp)
        >>> web_instance.terminate()
        >>> json_dict = resp.json()
        >>> cmdict_list = json_dict['response']
        >>> assert 'score_list' in cmdict_list[0]

    """
    kwargs['return_cm_simple_dict'] = True
    return ibs.query_chips(*args, **kwargs)


@register_ibs_method
@register_api('/api/query/chip/dict/', methods=['GET'])
def query_chips_dict(ibs, *args, **kwargs):
    """
    Runs query_chips, but returns a json compatible dictionary

    RESTful:
        Method: GET
        URL:    /api/query/chip/dict/
    """
    kwargs['return_cm_dict'] = True
    return ibs.query_chips(*args, **kwargs)


@register_ibs_method
@register_api('/api/review/query/graph/', methods=['POST'])
def process_graph_match_html(ibs, **kwargs):
    """
    RESTful:
        Method: POST
        URL:    /api/review/query/graph/
    """
    def sanitize(state):
        state = state.strip().lower()
        state = ''.join(state.split())
        return state
    import uuid
    map_dict = {
        'sameanimal'       : const.EVIDENCE_DECISION.INT_TO_CODE[const.EVIDENCE_DECISION.POSITIVE],
        'differentanimals' : const.EVIDENCE_DECISION.INT_TO_CODE[const.EVIDENCE_DECISION.NEGATIVE],
        'cannottell'       : const.EVIDENCE_DECISION.INT_TO_CODE[const.EVIDENCE_DECISION.INCOMPARABLE],
        'unreviewed'       : const.EVIDENCE_DECISION.INT_TO_CODE[const.EVIDENCE_DECISION.UNREVIEWED],
        'unknown'          : const.EVIDENCE_DECISION.INT_TO_CODE[const.EVIDENCE_DECISION.UNKNOWN],
        'photobomb'        : 'photobomb',
        'scenerymatch'     : 'scenerymatch',
        'excludetop'       : 'excludetop',
        'excludebottom'    : 'excludebottom',
    }
    annot_uuid_1 = uuid.UUID(request.form['identification-annot-uuid-1'])
    annot_uuid_2 = uuid.UUID(request.form['identification-annot-uuid-2'])
    state = request.form.get('identification-submit', '')
    state = sanitize(state)
    state = map_dict[state]
    tag_list = []
    if state in ['photobomb', 'scenerymatch']:
        tag_list.append(state)
        state = const.EVIDENCE_DECISION.NEGATIVE
    assert state in map_dict.values(), 'matching state is unrecognized'
    # Get checbox tags
    checbox_tag_list = ['photobomb', 'scenerymatch']
    for checbox_tag in checbox_tag_list:
        checkbox_name = 'ia-%s-value' % (checbox_tag)
        if checkbox_name in request.form:
            tag_list.append(checbox_tag)
    tag_list = sorted(set(tag_list))
    confidence_default = const.CONFIDENCE.INT_TO_CODE[const.CONFIDENCE.UNKNOWN]
    confidence = request.form.get('ia-turk-confidence', confidence_default)
    if confidence not in const.CONFIDENCE.CODE_TO_INT.keys():
        confidence = confidence_default
    if len(tag_list) == 0:
        tag_str = ''
    else:
        tag_str = ';'.join(tag_list)
    user_times = {
        'server_time_start' : request.form.get('server_time_start', None),
        'client_time_start' : request.form.get('client_time_start', None),
        'client_time_end'   : request.form.get('client_time_end',   None),
    }
    return (annot_uuid_1, annot_uuid_2, state, tag_str, 'web-api', confidence, user_times)


def ensure_review_image(ibs, aid, cm, qreq_, view_orientation='vertical',
                        draw_matches=True, verbose=False):
    r""""
    Create the review image for a pair of annotations

    CommandLine:
        python -m ibeis.web.apis_query ensure_review_image --show

    Example:
        >>> # SCRIPT
        >>> from ibeis.web.apis_query import *  # NOQA
        >>> import ibeis
        >>> cm, qreq_ = ibeis.testdata_cm('PZ_MTEST', a='default:dindex=0:10,qindex=0:1')
        >>> ibs = qreq_.ibs
        >>> aid = cm.get_top_aids()[0]
        >>> tt = ut.tic('make image')
        >>> image = ensure_review_image(ibs, aid, cm, qreq_)
        >>> ut.toc(tt)
        >>> ut.quit_if_noshow()
        >>> print('image.shape = %r' % (image.shape,))
        >>> print('image.dtype = %r' % (image.dtype,))
        >>> ut.print_object_size(image)
        >>> import plottool_ibeis as pt
        >>> pt.imshow(image)
        >>> ut.show_if_requested()
    """
    from ibeis.gui import id_review_api
    # Get thumb path
    match_thumb_path = ibs.get_match_thumbdir()
    match_thumb_filename = id_review_api.get_match_thumb_fname(cm, aid, qreq_,
                                                               view_orientation=view_orientation,
                                                               draw_matches=draw_matches)
    match_thumb_filepath = join(match_thumb_path, match_thumb_filename)
    if verbose:
        print('Checking: %r' % (match_thumb_filepath, ))

    if exists(match_thumb_filepath):
        image = cv2.imread(match_thumb_filepath)
    else:
        render_config = {
            'dpi'              : 150,
            'draw_fmatches'    : draw_matches,
            'vert'             : view_orientation == 'vertical',
            'show_aidstr'      : False,
            'show_name'        : False,
            'show_exemplar'    : False,
            'show_num_gt'      : False,
            'show_timedelta'   : False,
            'show_name_rank'   : False,
            'show_score'       : False,
            'show_annot_score' : False,
            'show_name_score'  : False,
            'draw_lbl'         : False,
            'draw_border'      : False,
        }

        if hasattr(qreq_, 'render_single_result'):
            image = qreq_.render_single_result(cm, aid, **render_config)
        else:
            image = cm.render_single_annotmatch(qreq_, aid, **render_config)
        #image = vt.crop_out_imgfill(image, fillval=(255, 255, 255), thresh=64)
        cv2.imwrite(match_thumb_filepath, image)
    return image


@register_api('/api/review/query/graph/alias/', methods=['POST'], __api_plural_check__=False)
def review_graph_match_html_alias(*args, **kwargs):
    review_graph_match_html(*args, **kwargs)


@register_api('/api/review/query/graph/', methods=['GET'])
def review_graph_match_html(ibs, review_pair, cm_dict, query_config_dict,
                            _internal_state, callback_url,
                            callback_method='POST',
                            view_orientation='vertical', include_jquery=False):
    r"""
    Args:
        ibs (ibeis.IBEISController):  image analysis api
        review_pair (dict): pair of annot uuids
        cm_dict (dict):
        query_config_dict (dict):
        _internal_state (?):
        callback_url (str):
        callback_method (unicode): (default = u'POST')
        view_orientation (unicode): (default = u'vertical')
        include_jquery (bool): (default = False)

    CommandLine:
        python -m ibeis.web.apis_query review_graph_match_html --show

        ibeis --web
        python -m ibeis.web.apis_query review_graph_match_html --show --domain=localhost

    Example:
        >>> # xdoctest: +REQUIRES(--web)
        >>> from ibeis.web.apis_query import *  # NOQA
        >>> import ibeis
        >>> web_ibs = ibeis.opendb_bg_web('testdb1')  # , domain='http://52.33.105.88')
        >>> aids = web_ibs.send_ibeis_request('/api/annot/', 'get')[0:2]
        >>> uuid_list = web_ibs.send_ibeis_request('/api/annot/uuid/', type_='get', aid_list=aids)
        >>> quuid_list = uuid_list[0:1]
        >>> duuid_list = uuid_list
        >>> query_config_dict = {
        >>>    # 'pipeline_root' : 'BC_DTW'
        >>> }
        >>> data = dict(
        >>>     query_annot_uuid_list=quuid_list, database_annot_uuid_list=duuid_list,
        >>>     query_config_dict=query_config_dict,
        >>> )
        >>> jobid = web_ibs.send_ibeis_request('/api/engine/query/graph/', **data)
        >>> print('jobid = %r' % (jobid,))
        >>> status_response = web_ibs.wait_for_results(jobid)
        >>> result_response = web_ibs.read_engine_results(jobid)
        >>> inference_result = result_response['json_result']
        >>> print('inference_result = %r' % (inference_result,))
        >>> auuid2_cm = inference_result['cm_dict']
        >>> quuid = quuid_list[0]
        >>> class_dict = auuid2_cm[str(quuid)]
        >>> # Get information in frontend
        >>> #ibs = ibeis.opendb('testdb1')
        >>> #cm = match_obj = ibeis.ChipMatch.from_dict(class_dict, ibs=ibs)
        >>> #match_obj.print_rawinfostr()
        >>> # Make the dictionary a bit more managable
        >>> #match_obj.compress_top_feature_matches(num=2)
        >>> #class_dict = match_obj.to_dict(ibs=ibs)
        >>> cm_dict = class_dict
        >>> # Package for review
        >>> review_pair = {'annot_uuid_1': quuid, 'annot_uuid_2': duuid_list[1]}
        >>> callback_method = u'POST'
        >>> view_orientation = u'vertical'
        >>> include_jquery = False
        >>> kw = dict(
        >>>     review_pair=review_pair,
        >>>     cm_dict=cm_dict,
        >>>     query_config_dict=query_config_dict,
        >>>     _internal_state=None,
        >>>     callback_url = None,
        >>> )
        >>> html_str = web_ibs.send_ibeis_request('/api/review/query/graph/', type_='get', **kw)
        >>> web_ibs.terminate2()
        >>> ut.quit_if_noshow()
        >>> import plottool_ibeis as pt
        >>> ut.render_html(html_str)
        >>> ut.show_if_requested()

    Example2:
        >>> # DISABLE_DOCTEST
        >>> # This starts off using web to get information, but finishes the rest in python
        >>> from ibeis.web.apis_query import *  # NOQA
        >>> import ibeis
        >>> ut.exec_funckw(review_graph_match_html, globals())
        >>> web_ibs = ibeis.opendb_bg_web('testdb1')  # , domain='http://52.33.105.88')
        >>> aids = web_ibs.send_ibeis_request('/api/annot/', 'get')[0:2]
        >>> uuid_list = web_ibs.send_ibeis_request('/api/annot/uuid/', type_='get', aid_list=aids)
        >>> quuid_list = uuid_list[0:1]
        >>> duuid_list = uuid_list
        >>> query_config_dict = {
        >>>    # 'pipeline_root' : 'BC_DTW'
        >>> }
        >>> data = dict(
        >>>     query_annot_uuid_list=quuid_list, database_annot_uuid_list=duuid_list,
        >>>     query_config_dict=query_config_dict,
        >>> )
        >>> jobid = web_ibs.send_ibeis_request('/api/engine/query/graph/', **data)
        >>> status_response = web_ibs.wait_for_results(jobid)
        >>> result_response = web_ibs.read_engine_results(jobid)
        >>> web_ibs.terminate2()
        >>> # NOW WORK IN THE FRONTEND
        >>> inference_result = result_response['json_result']
        >>> auuid2_cm = inference_result['cm_dict']
        >>> quuid = quuid_list[0]
        >>> class_dict = auuid2_cm[str(quuid)]
        >>> # Get information in frontend
        >>> ibs = ibeis.opendb('testdb1')
        >>> cm = ibeis.ChipMatch.from_dict(class_dict, ibs=ibs)
        >>> cm.print_rawinfostr()
        >>> # Make the dictionary a bit more managable
        >>> cm.compress_top_feature_matches(num=1)
        >>> cm.print_rawinfostr()
        >>> class_dict = cm.to_dict(ibs=ibs)
        >>> cm_dict = class_dict
        >>> # Package for review ( CANT CALL DIRECTLY BECAUSE OF OUT OF CONTEXT )
        >>> review_pair = {'annot_uuid_1': quuid, 'annot_uuid_2': duuid_list[1]}
        >>> x = review_graph_match_html(ibs, review_pair, cm_dict,
        >>>                             query_config_dict, _internal_state=None,
        >>>                             callback_url=None)
        >>> ut.quit_if_noshow()
        >>> import plottool_ibeis as pt
        >>> ut.render_html(html_str)
        >>> ut.show_if_requested()
    """
    from ibeis.algo.hots import chip_match
    # from ibeis.algo.hots.query_request import QueryRequest

    proot = query_config_dict.get('pipeline_root', 'vsmany')
    proot = query_config_dict.get('proot', proot)
    if proot.upper() in ('BC_DTW', 'OC_WDTW'):
        cls = chip_match.AnnotMatch  # ibs.depc_annot.requestclass_dict['BC_DTW']
    else:
        cls = chip_match.ChipMatch

    view_orientation = view_orientation.lower()
    if view_orientation not in ['vertical', 'horizontal']:
        view_orientation = 'horizontal'

    # unpack info
    try:
        annot_uuid_1 = review_pair['annot_uuid_1']
        annot_uuid_2 = review_pair['annot_uuid_2']
    except Exception:
        #??? HACK
        # FIXME:
        print('[!!!!] review_pair = %r' % (review_pair,))
        review_pair = review_pair[0]
        annot_uuid_1 = review_pair['annot_uuid_1']
        annot_uuid_2 = review_pair['annot_uuid_2']

    ibs.web_check_uuids(qannot_uuid_list=[annot_uuid_1],
                        dannot_uuid_list=[annot_uuid_2])

    aid_1 = ibs.get_annot_aids_from_uuid(annot_uuid_1)
    aid_2 = ibs.get_annot_aids_from_uuid(annot_uuid_2)

    cm = cls.from_dict(cm_dict, ibs=ibs)
    qreq_ = ibs.new_query_request([aid_1], [aid_2],
                                  cfgdict=query_config_dict)

    # Get score
    idx = cm.daid2_idx[aid_2]
    match_score = cm.name_score_list[idx]
    #match_score = cm.aid2_score[aid_2]

    try:
        image_matches = ensure_review_image(ibs, aid_2, cm, qreq_,
                                            view_orientation=view_orientation)
    except KeyError:
        image_matches = np.zeros((100, 100, 3), dtype=np.uint8)
        traceback.print_exc()
    try:
        image_clean = ensure_review_image(ibs, aid_2, cm, qreq_,
                                          view_orientation=view_orientation,
                                          draw_matches=False)
    except KeyError:
        image_clean = np.zeros((100, 100, 3), dtype=np.uint8)
        traceback.print_exc()

    image_matches_src = appf.embed_image_html(image_matches)
    image_clean_src = appf.embed_image_html(image_clean)

    confidence_dict = const.CONFIDENCE.NICE_TO_CODE
    confidence_nice_list = confidence_dict.keys()
    confidence_text_list = confidence_dict.values()
    confidence_selected_list = [
        confidence_text == 'unspecified'
        for confidence_text in confidence_text_list
    ]
    confidence_list = list(zip(confidence_nice_list, confidence_text_list, confidence_selected_list))

    if False:
        from ibeis.web import apis_query
        root_path = dirname(abspath(apis_query.__file__))
    else:
        root_path = dirname(abspath(__file__))
    css_file_list = [
        ['css', 'style.css'],
        ['include', 'bootstrap', 'css', 'bootstrap.css'],
    ]
    json_file_list = [
        ['javascript', 'script.js'],
        ['include', 'bootstrap', 'js', 'bootstrap.js'],
    ]

    if include_jquery:
        json_file_list = [
            ['javascript', 'jquery.min.js'],
        ] + json_file_list

    EMBEDDED_CSS = ''
    EMBEDDED_JAVASCRIPT = ''

    css_template_fmtstr = '<style type="text/css" ia-dependency="css">%s</style>\n'
    json_template_fmtstr = '<script type="text/javascript" ia-dependency="javascript">%s</script>\n'
    for css_file in css_file_list:
        css_filepath_list = [root_path, 'static'] + css_file
        with open(join(*css_filepath_list)) as css_file:
            EMBEDDED_CSS += css_template_fmtstr % (css_file.read(), )

    for json_file in json_file_list:
        json_filepath_list = [root_path, 'static'] + json_file
        with open(join(*json_filepath_list)) as json_file:
            EMBEDDED_JAVASCRIPT += json_template_fmtstr % (json_file.read(), )

    annot_uuid_1 = str(annot_uuid_1)
    annot_uuid_2 = str(annot_uuid_2)

    embedded = dict(globals(), **locals())
    return appf.template('turk', 'identification_insert', **embedded)


@register_route('/test/review/query/chip/', methods=['GET'])
def review_query_chips_test(**kwargs):
    """
    CommandLine:
        python -m ibeis.web.apis_query review_query_chips_test --show

    Example:
        >>> # SCRIPT
        >>> import ibeis
        >>> web_ibs = ibeis.opendb_bg_web(
        >>>     browser=True, url_suffix='/test/review/query/chip/?__format__=true')
    """
    ibs = current_app.ibs

    # the old block curvature dtw
    if 'use_bc_dtw' in request.args:
        query_config_dict = {
            'pipeline_root' : 'BC_DTW'
        }
    # the new oriented curvature dtw
    elif 'use_oc_wdtw' in request.args:
        query_config_dict = {
            'pipeline_root' : 'OC_WDTW'
        }
    else:
        query_config_dict = {}
    result_dict = ibs.query_chips_test(query_config_dict=query_config_dict)

    review_pair = result_dict['inference_dict']['annot_pair_dict']['review_pair_list'][0]
    annot_uuid_key = str(review_pair['annot_uuid_key'])
    cm_dict = result_dict['cm_dict'][annot_uuid_key]
    query_config_dict = result_dict['query_config_dict']
    _internal_state = result_dict['inference_dict']['_internal_state']
    callback_url = request.args.get('callback_url', url_for('process_graph_match_html'))
    callback_method = request.args.get('callback_method', 'POST')
    # view_orientation = request.args.get('view_orientation', 'vertical')
    view_orientation = request.args.get('view_orientation', 'horizontal')

    template_html = review_graph_match_html(ibs, review_pair, cm_dict,
                                            query_config_dict, _internal_state,
                                            callback_url, callback_method,
                                            view_orientation,
                                            include_jquery=True)
    template_html = '''
        <script src="http://code.jquery.com/jquery-2.2.1.min.js" ia-dependency="javascript"></script>
        %s
    ''' % (template_html, )
    return template_html
    return 'done'


@register_ibs_method
@register_api('/test/query/chip/', methods=['GET'])
def query_chips_test(ibs, **kwargs):
    """
    CommandLine:
        python -m ibeis.web.apis_query query_chips_test

    Example:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from ibeis.control.IBEISControl import *  # NOQA
        >>> import ibeis
        >>> qreq_ = ibeis.testdata_qreq_(defaultdb='testdb1')
        >>> ibs = qreq_.ibs
        >>> result_dict = ibs.query_chips_test()
        >>> print(result_dict)
    """
    from random import shuffle  # NOQA
    # Compile test data
    aid_list = ibs.get_valid_aids()
    # shuffle(aid_list)
    qaid_list = aid_list[:1]
    daid_list = aid_list[-4:]
    result_dict = ibs.query_chips_graph(qaid_list, daid_list, **kwargs)
    return result_dict


@register_ibs_method
@register_api('/api/query/graph/', methods=['GET', 'POST'])
def query_chips_graph(ibs, qaid_list, daid_list, user_feedback=None,
                      query_config_dict={}, echo_query_params=True):
    from ibeis.unstable.orig_graph_iden import OrigAnnotInference
    import theano  # NOQA
    import uuid

    def convert_to_uuid(nid):
        try:
            text = ibs.get_name_texts(nid)
            uuid_ = uuid.UUID(text)
        except ValueError:
            uuid_ = nid
        return uuid_

    cm_list, qreq_ = ibs.query_chips(qaid_list=qaid_list, daid_list=daid_list,
                                     cfgdict=query_config_dict, return_request=True)
    cm_dict = {
        str(ibs.get_annot_uuids(cm.qaid)): {
            # 'qaid'                  : cm.qaid,
            'qannot_uuid'           : ibs.get_annot_uuids(cm.qaid),
            # 'qnid'                  : cm.qnid,
            'qname_uuid'            : convert_to_uuid(cm.qnid),
            'qname'                 : ibs.get_name_texts(cm.qnid),
            # 'daid_list'             : cm.daid_list,
            'dannot_uuid_list'      : ibs.get_annot_uuids(cm.daid_list),
            # 'dnid_list'             : cm.dnid_list,
            'dname_uuid_list'       : [convert_to_uuid(nid) for nid in cm.dnid_list],
            # FIXME: use qreq_ state not ibeis state
            'dname_list'            : ibs.get_name_texts(cm.dnid_list),
            'score_list'            : cm.score_list,
            'annot_score_list'      : cm.annot_score_list,
            'fm_list'               : cm.fm_list if hasattr(cm, 'fm_list') else None,
            'fsv_list'              : cm.fsv_list if hasattr(cm, 'fsv_list') else None,
            # Non-corresponding lists to above
            # 'unique_nids'         : cm.unique_nids,
            'unique_name_uuid_list' : [convert_to_uuid(nid) for nid in cm.unique_nids],
            # FIXME: use qreq_ state not ibeis state
            'unique_name_list'      : ibs.get_name_texts(cm.unique_nids),
            'name_score_list'       : cm.name_score_list,
            # Placeholders for the reinitialization of the ChipMatch object
            'fk_list'               : None,
            'H_list'                : None,
            'fsv_col_lbls'          : None,
            'filtnorm_aids'         : None,
            'filtnorm_fxs'          : None,
        }
        for cm in cm_list
    }
    annot_inference = OrigAnnotInference(qreq_, cm_list, user_feedback)
    inference_dict = annot_inference.make_annot_inference_dict()
    result_dict = {
        'cm_dict'        : cm_dict,
        'inference_dict' : inference_dict,
    }
    if echo_query_params:
        result_dict['query_annot_uuid_list'] = ibs.get_annot_uuids(qaid_list)
        result_dict['database_annot_uuid_list'] = ibs.get_annot_uuids(daid_list)
        result_dict['query_config_dict'] = query_config_dict
    return result_dict


@register_ibs_method
@register_api('/api/query/chip/', methods=['GET'])
def query_chips(ibs, qaid_list=None, daid_list=None, cfgdict=None,
                use_cache=None, use_bigcache=None, qreq_=None,
                return_request=False, verbose=pipeline.VERB_PIPELINE,
                save_qcache=None, prog_hook=None, return_cm_dict=False,
                return_cm_simple_dict=False):
    r"""
    Submits a query request to the hotspotter recognition pipeline. Returns
    a list of QueryResult objects.

    Args:
        qaid_list (list): a list of annotation ids to be submitted as
            queries
        daid_list (list): a list of annotation ids used as the database
            that will be searched
        cfgdict (dict): dictionary of configuration options used to create
            a new QueryRequest if not already specified
        use_cache (bool): turns on/off chip match cache (default: True)
        use_bigcache (bool): turns one/off chunked chip match cache (default:
            True)
        qreq_ (QueryRequest): optional, a QueryRequest object that
            overrides all previous settings
        return_request (bool): returns the request which will be created if
            one is not already specified
        verbose (bool): default=False, turns on verbose printing

    Returns:
        list: a list of ChipMatch objects containing the matching
            annotations, scores, and feature matches

    Returns(2):
        tuple: (cm_list, qreq_) - a list of query results and optionally the
            QueryRequest object used

    RESTful:
        Method: PUT
        URL:    /api/query/chip/

    CommandLine:
        python -m ibeis.web.apis_query --test-query_chips

        # Test speed of single query
        python -m ibeis --tf IBEISController.query_chips --db PZ_Master1 \
            -a default:qindex=0:1,dindex=0:500 --nocache-hs

        python -m ibeis --tf IBEISController.query_chips --db PZ_Master1 \
            -a default:qindex=0:1,dindex=0:3000 --nocache-hs

        python -m ibeis.web.apis_query --test-query_chips:1 --show
        python -m ibeis.web.apis_query --test-query_chips:2 --show

    Example:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> from ibeis.control.IBEISControl import *  # NOQA
        >>> import ibeis
        >>> qreq_ = ibeis.testdata_qreq_()
        >>> ibs = qreq_.ibs
        >>> cm_list = qreq_.execute()
        >>> cm = cm_list[0]
        >>> ut.quit_if_noshow()
        >>> cm.ishow_analysis(qreq_)
        >>> ut.show_if_requested()

    Example:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> import ibeis
        >>> from ibeis.control.IBEISControl import *  # NOQA
        >>> qaid_list = [1]
        >>> daid_list = [1, 2, 3, 4, 5]
        >>> ibs = ibeis.opendb_test(db='testdb1')
        >>> qreq_ = ibs.new_query_request(qaid_list, daid_list)
        >>> cm = ibs.query_chips(qaid_list, daid_list, use_cache=False, qreq_=qreq_)[0]
        >>> ut.quit_if_noshow()
        >>> cm.ishow_analysis(qreq_)
        >>> ut.show_if_requested()

    Example1:
        >>> # SLOW_DOCTEST
        >>> # xdoctest: +SKIP
        >>> import ibeis
        >>> from ibeis.control.IBEISControl import *  # NOQA
        >>> qaid_list = [1]
        >>> daid_list = [1, 2, 3, 4, 5]
        >>> ibs = ibeis.opendb_test(db='testdb1')
        >>> cfgdict = {'pipeline_root':'BC_DTW'}
        >>> qreq_ = ibs.new_query_request(qaid_list, daid_list, cfgdict=cfgdict, verbose=True)
        >>> cm = ibs.query_chips(qreq_=qreq_)[0]
        >>> ut.quit_if_noshow()
        >>> cm.ishow_analysis(qreq_)
        >>> ut.show_if_requested()
    """
    # The qaid and daid objects are allowed to be None if qreq_ is
    # specified
    if qreq_ is None:
        assert qaid_list is not None, 'do not specify qaids and qreq'
        assert daid_list is not None, 'do not specify daids and qreq'
        qaid_list, was_scalar = ut.wrap_iterable(qaid_list)
        if daid_list is None:
            daid_list = ibs.get_valid_aids()
        qreq_ = ibs.new_query_request(qaid_list, daid_list,
                                      cfgdict=cfgdict, verbose=verbose)
    else:
        assert qaid_list is None, 'do not specify qreq and qaids'
        assert daid_list is None, 'do not specify qreq and daids'
        was_scalar = False
    cm_list = qreq_.execute()
    assert isinstance(cm_list, list), (
        'Chip matches were not returned as a list')

    # Convert to cm_list
    if return_cm_simple_dict:
        for cm in cm_list:
            cm.qauuid = ibs.get_annot_uuids(cm.qaid)
            cm.dauuid_list = ibs.get_annot_uuids(cm.daid_list)
        keys = ['qaid', 'daid_list', 'score_list', 'qauuid', 'dauuid_list']
        cm_list = [ut.dict_subset(cm.to_dict(), keys) for cm in cm_list]
    elif return_cm_dict:
        cm_list = [cm.to_dict() for cm in cm_list]

    if was_scalar:
        # hack for scalar input
        assert len(cm_list) == 1
        cm_list = cm_list[0]

    if return_request:
        return cm_list, qreq_
    else:
        return cm_list


##########################################################################################


@register_ibs_method
def get_graph_client_query_chips_graph_v2(ibs, graph_uuid):
    graph_client = current_app.GRAPH_CLIENT_DICT.get(graph_uuid, None)
    # We could be redirecting to a newer graph_client
    graph_uuid_chain = [graph_uuid]
    while isinstance(graph_client, six.string_types):
        graph_uuid_chain.append(graph_client)
        graph_client = current_app.GRAPH_CLIENT_DICT.get(graph_client, None)
    if graph_client is None:
        raise controller_inject.WebUnknownUUIDException(['graph_uuid'], [graph_uuid])
    return graph_client, graph_uuid_chain


def ensure_review_image_v2(ibs, match, draw_matches=False, draw_heatmask=False,
                           view_orientation='vertical', overlay=True):
    import plottool_ibeis as pt
    render_config = {
        'overlay'    : overlay,
        'show_ell'   : draw_matches,
        'show_lines' : draw_matches,
        'show_ori'   : False,
        'heatmask'   : draw_heatmask,
        'vert'       : view_orientation == 'vertical',
    }
    with pt.RenderingContext(dpi=150) as ctx:
        match.show(**render_config)
    image = ctx.image
    return image


def query_graph_v2_callback(graph_client, callback_type):
    from ibeis.web.graph_server import ut_to_json_encode
    assert callback_type in ['review', 'finished']
    callback_tuple = graph_client.callbacks.get(callback_type, None)
    if callback_tuple is not None:
        callback_url, callback_method = callback_tuple
        if callback_url is not None:
            callback_method = callback_method.lower()
            data_dict = ut_to_json_encode({
                'graph_uuid': graph_client.graph_uuid,
            })
            if callback_method == 'post':
                requests.post(callback_url, data=data_dict)
            elif callback_method == 'get':
                requests.get(callback_url, params=data_dict)
            elif callback_method == 'put':
                requests.put(callback_url, data=data_dict)
            elif callback_method == 'delete':
                requests.delete(callback_url, data=data_dict)
            else:
                raise KeyError('Unsupported HTTP callback method')


@register_ibs_method
@register_api('/api/query/graph/v2/', methods=['POST'])
def query_chips_graph_v2(ibs, annot_uuid_list=None,
                         query_config_dict={},
                         review_callback_url=None,
                         review_callback_method='POST',
                         finished_callback_url=None,
                         finished_callback_method='POST',
                         creation_imageset_rowid_list=None,
                         **kwargs):
    """
    CommandLine:
        python -m ibeis.web.apis_query --test-query_chips_graph_v2:0

        python -m ibeis reset_mtest_graph

        python -m ibeis --db PZ_MTEST --web --browser --url=/turk/identification/hardcase/
        python -m ibeis --db PZ_MTEST --web --browser --url=/turk/identification/graph/

    Example:
        >>> # xdoctest: +REQUIRES(--web)
        >>> from ibeis.web.apis_query import *
        >>> import ibeis
        >>> # Open local instance
        >>> ibs = ibeis.opendb('PZ_MTEST')
        >>> uuid_list = ibs.annots().uuids[0:10]
        >>> # Start up the web instance
        >>> web_ibs = ibeis.opendb_bg_web(db='PZ_MTEST', web=True, browser=False)
        >>> data = dict(annot_uuid_list=uuid_list)
        >>> resp = web_ibs.send_ibeis_request('/api/query/graph/v2/', **data)
        >>> print('resp = %r' % (resp,))
        >>> #cmdict_list = json_dict['response']
        >>> #assert 'score_list' in cmdict_list[0]

    Example:
        >>> # DEBUG_SCRIPT
        >>> from ibeis.web.apis_query import *
        >>> # Hack a flask context
        >>> current_app = ut.DynStruct()
        >>> current_app.GRAPH_CLIENT_DICT = {}
        >>> old = query_chips_graph_v2.__globals__.get('current_app', None)
        >>> query_chips_graph_v2.__globals__['current_app'] = current_app
        >>> import ibeis
        >>> ibs = ibeis.opendb('PZ_MTEST')
        >>> #ut.exec_funckw(query_chips_graph_v2, globals())
        >>> # Run function in main process
        >>> query_chips_graph_v2(ibs)
        >>> # Reset context
        >>> query_chips_graph_v2.__globals__['current_app'] = old
    """
    from ibeis.web.graph_server import GraphClient
    print('[apis_query] Creating GraphClient')

    if annot_uuid_list is None:
        annot_uuid_list = ibs.get_annot_uuids(ibs.get_valid_aids())

    ibs.web_check_uuids([], annot_uuid_list, [])
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)

    # FILTER FOR GGR2
    if True:
        aid_list = ibs.check_ggr_valid_aids(aid_list, **kwargs)

    graph_uuid = ut.hashable_to_uuid(sorted(aid_list))
    if graph_uuid not in current_app.GRAPH_CLIENT_DICT:
        for graph_uuid_ in current_app.GRAPH_CLIENT_DICT:
            graph_client_ = current_app.GRAPH_CLIENT_DICT[graph_uuid_]
            aid_list_ = graph_client_.aids
            assert aid_list_ is not None
            overlap_aid_set = set(aid_list_) & set(aid_list)
            if len(overlap_aid_set) > 0:
                overlap_aid_list = list(overlap_aid_set)
                overlap_annot_uuid_list = ibs.get_annot_uuids(overlap_aid_list)
                raise controller_inject.WebUnavailableUUIDException(
                    overlap_annot_uuid_list, graph_uuid_)

        callback_dict = {
            'review'   : (review_callback_url,   review_callback_method),
            'finished' : (finished_callback_url, finished_callback_method),
        }
        graph_client = GraphClient(graph_uuid, callbacks=callback_dict,
                                   autoinit=True)

        if creation_imageset_rowid_list is not None:
            graph_client.imagesets = creation_imageset_rowid_list
        graph_client.aids = aid_list

        config = {
            'manual.n_peek'   : GRAPH_CLIENT_PEEK,
            'manual.autosave' : True,
            'redun.pos'       : 2,
            'redun.neg'       : 2,
            'algo.quickstart' : False
        }
        config.update(query_config_dict)
        print('[apis_query] graph_client.config = {}'.format(ut.repr3(config)))
        graph_client.config = config

        # Ensure no race-conditions
        current_app.GRAPH_CLIENT_DICT[graph_uuid] = graph_client

        # Start (create the Graph Inference object)
        payload = {
            'action' : 'start',
            'dbdir'  : ibs.dbdir,
            'aids'   : graph_client.aids,
            'config' : graph_client.config,
        }
        future = graph_client.post(payload)
        future.result()  # Guarantee that this has happened before calling refresh

        f2 = graph_client.post({'action' : 'latest_logs'})
        f2.graph_client = graph_client
        f2.add_done_callback(query_graph_v2_latest_logs)

        # Start (create the Graph Inference object)
        payload = {
            'action' : 'get_feat_extractor',
        }
        future = graph_client.post(payload)
        graph_client.extr = future.result()

        # Start main loop
        future = graph_client.post({'action' : 'continue_review'})
        future.graph_client = graph_client
        future.add_done_callback(query_graph_v2_on_request_review)

        f2 = graph_client.post({'action' : 'latest_logs'})
        f2.graph_client = graph_client
        f2.add_done_callback(query_graph_v2_latest_logs)

    return graph_uuid


@register_ibs_method
def review_graph_match_config_v2(ibs, graph_uuid, aid1=None, aid2=None,
                                 view_orientation='vertical', view_version=1):
    from ibeis.algo.verif import pairfeat
    from flask import session

    EDGES_KEY = '_EDGES_'
    EDGES_MAX = 10

    user_id = controller_inject.get_user()
    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)

    if aid1 is not None and aid2 is not None:
        previous_edge_list = None
        if aid1 > aid2:
            aid1, aid2 = aid2, aid1
        edge = (aid1, aid2)
        data = graph_client.check(edge)
        if data is None:
            data = (
                edge,
                np.nan,
                {},
            )
    else:
        if EDGES_KEY not in session:
            session[EDGES_KEY] = []
        previous_edge_list = session[EDGES_KEY]
        print('Using previous_edge_list\n\tUser: %s\n\tList: %r' % (user_id, previous_edge_list, ))

        data = graph_client.sample(previous_edge_list=previous_edge_list, max_previous_edges=EDGES_MAX)
        if data is None:
            raise controller_inject.WebReviewNotReadyException(graph_uuid)

    edge, priority, data_dict = data

    edge_ = [
        int(edge[0]),
        int(edge[1]),
    ]
    if previous_edge_list is not None:
        previous_edge_list.append(edge_)
        if len(previous_edge_list) > EDGES_MAX:
            cutoff = int(-1.0 * EDGES_MAX)
            previous_edge_list = previous_edge_list[cutoff:]
        session[EDGES_KEY] = previous_edge_list
        print('Updating previous_edge_list\n\tUser: %s\n\tList: %r' % (user_id, previous_edge_list, ))

    args = (edge, priority, )
    print('Sampled edge %r with priority %0.02f' % args)
    print('Data: ' + ut.repr4(data_dict))

    aid_1, aid_2 = edge
    annot_uuid_1 = str(ibs.get_annot_uuids(aid_1))
    annot_uuid_2 = str(ibs.get_annot_uuids(aid_2))

    feat_extract_config = {
        'match_config': ({} if graph_client.extr is None else
                         graph_client.extr.match_config)
    }
    extr = pairfeat.PairwiseFeatureExtractor(ibs, config=feat_extract_config)

    match = extr._exec_pairwise_match([edge])[0]

    image_clean = ensure_review_image_v2(ibs, match,
                                         view_orientation=view_orientation,
                                         overlay=False)
    # image_matches = ensure_review_image_v2(ibs, match, draw_matches=True,
    #                                        view_orientation=view_orientation)

    print('Using View Version: %r' % (view_version, ))
    if view_version == 1:
        image_heatmask = ensure_review_image_v2(ibs, match, draw_heatmask=True,
                                                view_orientation=view_orientation)
    else:
        image_heatmask = ensure_review_image_v2(ibs, match, draw_matches=True,
                                                view_orientation=view_orientation)

    image_clean_src = appf.embed_image_html(image_clean)
    # image_matches_src = appf.embed_image_html(image_matches)
    image_heatmask_src = appf.embed_image_html(image_heatmask)

    now = datetime.utcnow()
    server_time_start = float(now.strftime("%s.%f"))

    return (edge, priority, data_dict, aid_1, aid_2, annot_uuid_1, annot_uuid_2,
            image_clean_src, image_heatmask_src, image_heatmask_src,
            server_time_start)


@register_api('/api/review/query/graph/v2/', methods=['GET'])
def review_graph_match_html_v2(ibs, graph_uuid, callback_url=None,
                               callback_method='POST',
                               view_orientation='vertical',
                               view_version=1,
                               include_jquery=False):
    values = ibs.review_graph_match_config_v2(graph_uuid,
                                              view_orientation=view_orientation,
                                              view_version=view_version)

    (edge, priority, data_dict, aid1, aid2, annot_uuid_1, annot_uuid_2,
        image_clean_src, image_matches_src, image_heatmask_src,
        server_time_start) = values

    confidence_dict = const.CONFIDENCE.NICE_TO_CODE
    confidence_nice_list = confidence_dict.keys()
    confidence_text_list = confidence_dict.values()
    confidence_selected_list = [
        confidence_text == 'unspecified'
        for confidence_text in confidence_text_list
    ]
    confidence_list = list(zip(confidence_nice_list, confidence_text_list, confidence_selected_list))

    if False:
        from ibeis.web import apis_query
        root_path = dirname(abspath(apis_query.__file__))
    else:
        root_path = dirname(abspath(__file__))
    css_file_list = [
        ['css', 'style.css'],
        ['include', 'bootstrap', 'css', 'bootstrap.css'],
    ]
    json_file_list = [
        ['javascript', 'script.js'],
        ['include', 'bootstrap', 'js', 'bootstrap.js'],
    ]

    if include_jquery:
        json_file_list = [
            ['javascript', 'jquery.min.js'],
        ] + json_file_list

    EMBEDDED_CSS = ''
    EMBEDDED_JAVASCRIPT = ''

    css_template_fmtstr = '<style type="text/css" ia-dependency="css">%s</style>\n'
    json_template_fmtstr = '<script type="text/javascript" ia-dependency="javascript">%s</script>\n'
    for css_file in css_file_list:
        css_filepath_list = [root_path, 'static'] + css_file
        with open(join(*css_filepath_list)) as css_file:
            EMBEDDED_CSS += css_template_fmtstr % (css_file.read(), )

    for json_file in json_file_list:
        json_filepath_list = [root_path, 'static'] + json_file
        with open(join(*json_filepath_list)) as json_file:
            EMBEDDED_JAVASCRIPT += json_template_fmtstr % (json_file.read(), )

    embedded = dict(globals(), **locals())
    return appf.template('turk', 'identification_insert', **embedded)


@register_api('/api/status/query/graph/v2/', methods=['GET'], __api_plural_check__=False)
def view_graphs_status(ibs):
    graph_dict = {}
    for graph_uuid in current_app.GRAPH_CLIENT_DICT:
        graph_client = current_app.GRAPH_CLIENT_DICT.get(graph_uuid, None)
        if graph_client is None:
            continue
        graph_status, graph_exception = graph_client.refresh_status()
        if graph_client.review_dict is None:
            num_edges = None
        else:
            edge_list = list(graph_client.review_dict.keys())
            num_edges = len(edge_list)
        graph_uuid = str(graph_uuid)
        graph_dict[graph_uuid] = {
            'status': graph_status,
            'num_aids': len(graph_client.aids),
            'num_reviews': num_edges,
        }
    return graph_dict


@register_ibs_method
@register_api('/api/review/query/graph/v2/', methods=['POST'])
def process_graph_match_html_v2(ibs, graph_uuid, **kwargs):
    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)
    response_tuple = process_graph_match_html(ibs, **kwargs)
    annot_uuid_1, annot_uuid_2, decision, tags, user_id, confidence, user_times = response_tuple
    aid1 = ibs.get_annot_aids_from_uuid(annot_uuid_1)
    aid2 = ibs.get_annot_aids_from_uuid(annot_uuid_2)
    edge = (aid1, aid2, )
    user_id = controller_inject.get_user()
    now = datetime.utcnow()

    if decision in ['excludetop', 'excludebottom']:
        aid = aid1 if decision == 'excludetop' else aid2

        metadata_dict = ibs.get_annot_metadata(aid)
        assert 'excluded' not in metadata_dict
        metadata_dict['excluded'] = True
        ibs.set_annot_metadata([aid], [metadata_dict])

        payload = {
            'action'            : 'remove_annots',
            'aids'              : [aid],
        }
    else:
        payload = {
            'action'            : 'add_feedback',
            'edge'              : edge,
            'evidence_decision' : decision,
            # TODO: meta_decision should come from the html resp.  When generating
            # the html page, the default value should be its previous value. If the
            # user changes it to be something incompatible them perhaps just reset
            # it to null.
            'meta_decision'     : 'null',
            'tags'              : [] if len(tags) == 0 else tags.split(';'),
            'user_id'           : 'user:web:%s' % (user_id, ),
            'confidence'        : confidence,
            'timestamp_s1'      : user_times['server_time_start'],
            'timestamp_c1'      : user_times['client_time_start'],
            'timestamp_c2'      : user_times['client_time_end'],
            'timestamp'         : float(now.strftime("%s.%f"))
        }
    print('POSTING GRAPH CLIENT REVIEW:')
    print(ut.repr4(payload))
    graph_client.post(payload)

    # Clean any old continue_reviews
    graph_client.cleanup()

    # Continue review
    future = graph_client.post({'action' : 'continue_review'})
    future.graph_client = graph_client
    future.add_done_callback(query_graph_v2_on_request_review)

    f2 = graph_client.post({'action' : 'latest_logs'})
    f2.graph_client = graph_client
    f2.add_done_callback(query_graph_v2_latest_logs)
    return (annot_uuid_1, annot_uuid_2, )


@register_ibs_method
@register_api('/api/query/graph/v2/', methods=['GET'])
def sync_query_chips_graph_v2(ibs, graph_uuid):
    import ibeis
    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)

    # Create the AnnotInference
    infr = ibeis.AnnotInference(ibs=ibs, aids=graph_client.aids, autoinit=True)
    for key in graph_client.config:
        infr.params[key] = graph_client.config[key]
    infr.reset_feedback('staging', apply=True)

    infr.relabel_using_reviews(rectify=True)
    edge_delta_df = infr.match_state_delta(old='annotmatch', new='all')
    name_delta_df = infr.get_ibeis_name_delta()

    ############################################################################

    col_list = list(edge_delta_df.columns)
    match_aid_edge_list = list(edge_delta_df.index)
    match_aid1_list = ut.take_column(match_aid_edge_list, 0)
    match_aid2_list = ut.take_column(match_aid_edge_list, 1)
    match_annot_uuid1_list = ibs.get_annot_uuids(match_aid1_list)
    match_annot_uuid2_list = ibs.get_annot_uuids(match_aid2_list)
    match_annot_uuid_edge_list = list(zip(match_annot_uuid1_list, match_annot_uuid2_list))

    zipped = list(zip(*( list(edge_delta_df[col]) for col in col_list )))

    match_list = []
    for match_annot_uuid_edge, zipped_ in list(zip(match_annot_uuid_edge_list, zipped)):
        match_dict = {
            'edge': match_annot_uuid_edge,
        }
        for index, col in enumerate(col_list):
            match_dict[col] = zipped_[index]
        match_list.append(match_dict)

    ############################################################################

    col_list = list(name_delta_df.columns)
    name_aid_list = list(name_delta_df.index)
    name_annot_uuid_list = ibs.get_annot_uuids(name_aid_list)
    old_name_list = list(name_delta_df['old_name'])
    new_name_list = list(name_delta_df['new_name'])
    zipped = list(zip(name_annot_uuid_list, old_name_list, new_name_list))
    name_dict = {
        str(name_annot_uuid): {
            'old': old_name,
            'new': new_name,
        }
        for name_annot_uuid, old_name, new_name in zipped
    }

    ############################################################################

    ret_dict = {
        'match_list'  : match_list,
        'name_dict'   : name_dict,
    }

    infr.write_ibeis_staging_feedback()
    infr.write_ibeis_annotmatch_feedback(edge_delta_df)
    infr.write_ibeis_name_assignment(name_delta_df)
    edge_delta_df.reset_index()

    return ret_dict


@register_ibs_method
@register_api('/api/query/graph/v2/', methods=['PUT'])
def add_annots_query_chips_graph_v2(ibs, graph_uuid, annot_uuid_list):
    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)
    ibs.web_check_uuids([], annot_uuid_list, [])
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)

    for graph_uuid_ in current_app.GRAPH_CLIENT_DICT:
        graph_client_ = current_app.GRAPH_CLIENT_DICT[graph_uuid_]
        aid_list_ = graph_client_.aids
        assert aid_list_ is not None
        overlap_aid_set = set(aid_list_) & set(aid_list)
        if len(overlap_aid_set) > 0:
            overlap_aid_list = list(overlap_aid_set)
            overlap_annot_uuid_list = ibs.get_annot_uuids(overlap_aid_list)
            raise controller_inject.WebUnavailableUUIDException(
                overlap_annot_uuid_list, graph_uuid_)

    aid_list_ = graph_client.aids + aid_list
    graph_uuid_ = ut.hashable_to_uuid(sorted(aid_list_))
    assert graph_uuid_ not in current_app.GRAPH_CLIENT_DICT
    graph_client.graph_uuid = graph_uuid_

    payload = {
        'action' : 'add_annots',
        'dbdir'  : ibs.dbdir,
        'aids'   : aid_list,
    }
    future = graph_client.post(payload)
    future.result()  # Guarantee that this has happened before calling refresh

    # Start main loop
    future = graph_client.post({'action' : 'continue_review'})
    future.graph_client = graph_client
    future.add_done_callback(query_graph_v2_on_request_review)

    current_app.GRAPH_CLIENT_DICT[graph_uuid_] = graph_client
    current_app.GRAPH_CLIENT_DICT[graph_uuid] = graph_uuid_
    return graph_uuid_


@register_ibs_method
def remove_annots_query_chips_graph_v2(ibs, graph_uuid, annot_uuid_list):
    graph_client, _ = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)
    ibs.web_check_uuids([], annot_uuid_list, [])
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)

    aid_list_ = list(set(graph_client.aids) - set(aid_list))
    graph_uuid_ = ut.hashable_to_uuid(sorted(aid_list_))
    assert graph_uuid_ not in current_app.GRAPH_CLIENT_DICT
    graph_client.graph_uuid = graph_uuid_

    payload = {
        'action' : 'remove_annots',
        'dbdir'  : ibs.dbdir,
        'aids'   : aid_list,
    }
    future = graph_client.post(payload)
    future.result()  # Guarantee that this has happened before calling refresh

    # Start main loop
    future = graph_client.post({'action' : 'continue_review'})
    future.graph_client = graph_client
    future.add_done_callback(query_graph_v2_on_request_review)

    current_app.GRAPH_CLIENT_DICT[graph_uuid_] = graph_client
    current_app.GRAPH_CLIENT_DICT[graph_uuid] = graph_uuid_
    return graph_uuid_


@register_ibs_method
@register_api('/api/query/graph/v2/', methods=['DELETE'])
def delete_query_chips_graph_v2(ibs, graph_uuid):
    values = ibs.get_graph_client_query_chips_graph_v2(graph_uuid)
    graph_client, graph_uuid_chain = values
    del graph_client
    for graph_uuid_ in graph_uuid_chain:
        if graph_uuid_ in current_app.GRAPH_CLIENT_DICT:
            current_app.GRAPH_CLIENT_DICT[graph_uuid_] = None
            current_app.GRAPH_CLIENT_DICT.pop(graph_uuid_)
    return True


def query_graph_v2_latest_logs(future):
    if not future.cancelled():
        logs = future.result()
        print('--- <LOG DUMP> ---')
        for msg, color in logs:
            ut.cprint('[web.infr] ' + msg, color)
        print('--- <\LOG DUMP> ---')


def query_graph_v2_on_request_review(future):
    if not future.cancelled():
        graph_client = future.graph_client
        data_list = future.result()
        if data_list is not None:
            graph_client.update(data_list)
            callback_type = 'review'
        else:
            graph_client.update(None)
            callback_type = 'finished'
        query_graph_v2_callback(graph_client, callback_type)


if __name__ == '__main__':
    """
    CommandLine:
        python -m ibeis.web.app
        python -m ibeis.web.app --allexamples
        python -m ibeis.web.app --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
Dependencies: flask, tornado
"""
from __future__ import absolute_import, division, print_function
from flask import request, redirect, url_for, current_app
from ibeis.control import controller_inject
from ibeis.web import appfuncs as appf
from ibeis import constants as const
import utool as ut
import numpy as np
import uuid
import six


register_route = controller_inject.get_ibeis_flask_route(__name__)


@register_route('/submit/login/', methods=['POST'], __route_authenticate__=False)
def submit_login(name, organization, refer=None, *args, **kwargs):
    # Return HTML
    if refer is None:
        refer = url_for('root')
    else:
        refer = appf.decode_refer_url(refer)

    if name == '_new_':
        first = kwargs['new_name_first']
        last = kwargs['new_name_last']
        name = '%s.%s' % (first, last, )
        name = name.replace(' ', '')

    if organization == '_new_':
        organization = kwargs['new_organization']
        organization = organization.replace(' ', '')

    name = name.lower()
    organization = organization.lower()

    username = '%s@%s' % (name, organization, )
    controller_inject.authenticate(username=username, name=name,
                                   organization=organization)

    return redirect(refer)


@register_route('/submit/cameratrap/', methods=['POST'])
def submit_cameratrap(**kwargs):
    ibs = current_app.ibs
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    gid = int(request.form['cameratrap-gid'])
    user_id = controller_inject.get_user().get('username', None)
    flag = request.form.get('cameratrap-toggle', 'off') == 'on'
    ibs.set_image_cameratrap([gid], [flag])
    print('[web] user_id: %s, gid: %d, flag: %r' % (user_id, gid, flag, ))

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_cameratrap', imgsetid=imgsetid, previous=gid))


@register_route('/submit/detection/', methods=['POST'])
def submit_detection(**kwargs):
    is_staged = kwargs.get('staged', False)

    ibs = current_app.ibs
    method = request.form.get('detection-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)
    gid = int(request.form['detection-gid'])
    user_id = controller_inject.get_user().get('username', None)

    poor_boxes = method.lower() == 'poor boxes'
    if poor_boxes:
        imgsetid_ = ibs.get_imageset_imgsetids_from_text('POOR BOXES')
        ibs.set_image_imgsetids([gid], [imgsetid_])
        method = 'accept'

    if method.lower() == 'delete':
        # ibs.delete_images(gid)
        # print('[web] (DELETED) user_id: %s, gid: %d' % (user_id, gid, ))
        pass
    elif method.lower() == 'clear':
        aid_list = ibs.get_image_aids(gid)
        ibs.delete_annots(aid_list)
        print('[web] (CLEAERED) user_id: %s, gid: %d' % (user_id, gid, ))
        redirection = request.referrer
        if 'gid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&gid=%d' % (redirection, gid, )
            else:
                redirection = '%s?gid=%d' % (redirection, gid, )
        return redirect(redirection)
    elif method.lower() == 'rotate left':
        ibs.update_image_rotate_left_90([gid])
        print('[web] (ROTATED LEFT) user_id: %s, gid: %d' % (user_id, gid, ))
        redirection = request.referrer
        if 'gid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&gid=%d' % (redirection, gid, )
            else:
                redirection = '%s?gid=%d' % (redirection, gid, )
        return redirect(redirection)
    elif method.lower() == 'rotate right':
        ibs.update_image_rotate_right_90([gid])
        print('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, gid, ))
        redirection = request.referrer
        if 'gid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&gid=%d' % (redirection, gid, )
            else:
                redirection = '%s?gid=%d' % (redirection, gid, )
        return redirect(redirection)
    else:
        current_aid_list = ibs.get_image_aids(gid, is_staged=is_staged)
        current_part_rowid_list = ut.flatten(ibs.get_annot_part_rowids(current_aid_list, is_staged=is_staged))

        if is_staged:
            staged_uuid = uuid.uuid4()
            staged_user = controller_inject.get_user()
            staged_user_id = staged_user.get('username', None)

            # Filter aids for current user
            current_annot_user_id_list = ibs.get_annot_staged_user_ids(current_aid_list)
            current_aid_list = [
                current_aid
                for current_aid, current_annot_user_id in zip(current_aid_list, current_annot_user_id_list)
                if current_annot_user_id == staged_user_id
            ]

            # Filter part_rowids for current user
            current_part_user_id_list = ibs.get_part_staged_user_ids(current_part_rowid_list)
            current_part_rowid_list = [
                current_part_rowid
                for current_part_rowid, current_part_user_id in zip(current_part_rowid_list, current_part_user_id_list)
                if current_part_user_id == staged_user_id
            ]
        else:
            staged_uuid = None
            staged_user = None
            staged_user_id = None

        # Make new annotations
        width, height = ibs.get_image_sizes(gid)

        # Separate out annotations vs parts
        data_list = ut.from_json(request.form['ia-detection-data'])
        print(request.form['ia-detection-manifest'])
        raw_manifest = request.form['ia-detection-manifest'].strip()
        try:
            manifest_list = ut.from_json(raw_manifest)
        except ValueError:
            manifest_list = []
        test_truth = len(manifest_list) > 0
        test_challenge_list = [{
            'gid'           : gid,
            'manifest_list' : manifest_list,
        }]
        test_response_list = [{
            'poor_boxes'    : poor_boxes,
        }]
        test_result_list = [test_truth == poor_boxes]
        test_user_id_list = [None]
        ibs.add_test(test_challenge_list, test_response_list,
                     test_result_list=test_result_list,
                     test_user_identity_list=test_user_id_list)

        if not test_truth:
            annotation_list = []
            part_list = []
            mapping_dict = {}
            for outer_index, data in enumerate(data_list):
                # Check for invalid NaN boxes, filter them out
                try:
                    assert data['percent']['left']   is not None
                    assert data['percent']['top']    is not None
                    assert data['percent']['width']  is not None
                    assert data['percent']['height'] is not None
                except:
                    continue

                parent_index = data['parent']
                if parent_index is None:
                    inner_index = len(annotation_list)
                    annotation_list.append(data)
                    mapping_dict[outer_index] = inner_index
                else:
                    assert parent_index in mapping_dict
                    part_list.append(data)

            ##################################################################################
            # Get primatives
            bbox_list = [
                (
                    int(np.round(width  * (annot['percent']['left']   / 100.0) )),
                    int(np.round(height * (annot['percent']['top']    / 100.0) )),
                    int(np.round(width  * (annot['percent']['width']  / 100.0) )),
                    int(np.round(height * (annot['percent']['height'] / 100.0) )),
                )
                for annot in annotation_list
            ]
            theta_list = [
                float(annot['angles']['theta'])
                for annot in annotation_list
            ]

            # Get metadata
            viewpoint1_list = [
                int(annot['metadata'].get('viewpoint1', -1))
                for annot in annotation_list
            ]
            viewpoint2_list = [
                int(annot['metadata'].get('viewpoint2', -1))
                for annot in annotation_list
            ]
            viewpoint3_list = [
                int(annot['metadata'].get('viewpoint3', -1))
                for annot in annotation_list
            ]
            zipped = zip(viewpoint1_list, viewpoint2_list, viewpoint3_list)
            viewpoint_list = [ appf.convert_tuple_to_viewpoint(tup) for tup in zipped ]

            quality_list = [
                int(annot['metadata'].get('quality', 0))
                for annot in annotation_list
            ]
            # Fix qualities
            for index, quality in enumerate(quality_list):
                if quality == 0:
                    quality_list[index] = None
                elif quality == 1:
                    quality_list[index] = 2
                elif quality == 2:
                    quality_list[index] = 4
                else:
                    raise ValueError('quality must be 0, 1 or 2')

            multiple_list =  [
                annot['metadata'].get('multiple', False)
                for annot in annotation_list
            ]
            interest_list =  [
                annot['highlighted']
                for annot in annotation_list
            ]
            species_list = [
                annot['metadata'].get('species', const.UNKNOWN)
                for annot in annotation_list
            ]

            # Process annotations
            survived_aid_list = [
                None if annot['label'] in [None, 'None'] else int(annot['label'])
                for annot in annotation_list
            ]

            # Delete annotations that didn't survive
            kill_aid_list = list(set(current_aid_list) - set(survived_aid_list))
            ibs.delete_annots(kill_aid_list)

            staged_uuid_list = [staged_uuid] * len(survived_aid_list)
            staged_user_id_list = [staged_user_id] * len(survived_aid_list)

            aid_list = []
            zipped = zip(survived_aid_list, bbox_list, staged_uuid_list, staged_user_id_list)
            for aid, bbox, staged_uuid, staged_user_id in zipped:
                staged_uuid_list_ = None if staged_uuid is None else [staged_uuid]
                staged_user_id_list_ = None if staged_user_id is None else [staged_user_id]

                if aid is None:
                    aid_ = ibs.add_annots([gid], [bbox],
                                          staged_uuid_list=staged_uuid_list_,
                                          staged_user_id_list=staged_user_id_list_)
                    aid_ = aid_[0]
                else:
                    ibs.set_annot_bboxes([aid], [bbox])
                    if staged_uuid_list_ is not None:
                        ibs.set_annot_staged_uuids([aid], staged_uuid_list_)
                    if staged_user_id_list_ is not None:
                        ibs.set_annot_staged_user_ids([aid], staged_user_id_list_)
                    aid_ = aid
                aid_list.append(aid_)

            print('aid_list = %r' % (aid_list, ))
            # Set annotation metadata
            ibs.set_annot_thetas(aid_list, theta_list)
            ibs.set_annot_viewpoints(aid_list, viewpoint_list)
            # TODO ibs.set_annot_viewpoint_code(aid_list, viewpoint_list)
            ibs.set_annot_qualities(aid_list, quality_list)
            ibs.set_annot_multiple(aid_list, multiple_list)
            ibs.set_annot_interest(aid_list, interest_list)
            ibs.set_annot_species(aid_list, species_list)

            # Set the mapping dict to use aids now
            mapping_dict = { key: aid_list[index] for key, index in mapping_dict.items() }

            ##################################################################################
            # Process parts
            survived_part_rowid_list = [
                None if part['label'] is None else int(part['label'])
                for part in part_list
            ]

            # Get primatives
            aid_list = [
                mapping_dict[part['parent']]
                for part in part_list
            ]
            bbox_list = [
                (
                    int(np.round(width  * (part['percent']['left']   / 100.0) )),
                    int(np.round(height * (part['percent']['top']    / 100.0) )),
                    int(np.round(width  * (part['percent']['width']  / 100.0) )),
                    int(np.round(height * (part['percent']['height'] / 100.0) )),
                )
                for part in part_list
            ]
            theta_list = [
                float(part['angles']['theta'])
                for part in part_list
            ]

            # Get metadata
            viewpoint1_list = [
                int(part['metadata'].get('viewpoint1', -1))
                for part in part_list
            ]
            viewpoint2_list = [-1] * len(part_list)
            viewpoint3_list = [-1] * len(part_list)
            zipped = zip(viewpoint1_list, viewpoint2_list, viewpoint3_list)
            viewpoint_list = [ appf.convert_tuple_to_viewpoint(tup) for tup in zipped ]

            quality_list = [
                int(part['metadata'].get('quality', 0))
                for part in part_list
            ]
            # Fix qualities
            for index, quality in enumerate(quality_list):
                if quality == 0:
                    quality_list[index] = None
                elif quality == 1:
                    quality_list[index] = 2
                elif quality == 2:
                    quality_list[index] = 4
                else:
                    raise ValueError('quality must be 0, 1 or 2')

            type_list = [
                part['metadata'].get('type', const.UNKNOWN)
                for part in part_list
            ]

            # Delete annotations that didn't survive
            kill_part_rowid_list = list(set(current_part_rowid_list) - set(survived_part_rowid_list))
            ibs.delete_parts(kill_part_rowid_list)

            staged_uuid_list = [staged_uuid] * len(survived_part_rowid_list)
            staged_user_id_list = [staged_user_id] * len(survived_part_rowid_list)

            part_rowid_list = []
            zipped = zip(survived_part_rowid_list, aid_list, bbox_list, staged_uuid_list, staged_user_id_list)
            for part_rowid, aid, bbox, staged_uuid, staged_user_id in zipped:
                staged_uuid_list_ = None if staged_uuid is None else [staged_uuid]
                staged_user_id_list_ = None if staged_user_id is None else [staged_user_id]

                if part_rowid is None:
                    part_rowid_ = ibs.add_parts([aid], [bbox],
                                                staged_uuid_list=staged_uuid_list_,
                                                staged_user_id_list=staged_user_id_list_)
                    part_rowid_ = part_rowid_[0]
                else:
                    ibs._set_part_aid([part_rowid], [aid])
                    ibs.set_part_bboxes([part_rowid], [bbox])
                    if staged_uuid_list_ is not None:
                        ibs.set_part_staged_uuids([part_rowid], staged_uuid_list_)
                    if staged_user_id_list_ is not None:
                        ibs.set_part_staged_user_ids([part_rowid], staged_user_id_list_)

                    part_rowid_ = part_rowid
                part_rowid_list.append(part_rowid_)

            # Set part metadata
            print('part_rowid_list = %r' % (part_rowid_list, ))
            ibs.set_part_thetas(part_rowid_list, theta_list)
            ibs.set_part_viewpoints(part_rowid_list, viewpoint_list)
            ibs.set_part_qualities(part_rowid_list, quality_list)
            ibs.set_part_types(part_rowid_list, type_list)

            # Set image reviewed flag
            if is_staged:
                metadata_dict = ibs.get_image_metadata(gid)
                if 'staged' not in metadata_dict:
                    metadata_dict['staged'] = {
                        'sessions': {
                            'uuids': [],
                            'user_ids': [],
                        }
                    }
                metadata_dict['staged']['sessions']['uuids'].append(str(staged_uuid))
                metadata_dict['staged']['sessions']['user_ids'].append(staged_user_id)
                ibs.set_image_metadata([gid], [metadata_dict])
            else:
                ibs.set_image_reviewed([gid], [1])

            print('[web] user_id: %s, gid: %d, annots: %d, parts: %d' % (user_id, gid, len(annotation_list), len(part_list), ))

    default_list = [
        'autointerest',
        'interest_bypass',
        'metadata',
        'metadata_viewpoint',
        'metadata_quality',
        'metadata_flags',
        'metadata_flags_aoi',
        'metadata_flags_multiple',
        'metadata_species',
        'metadata_label',
        'metadata_quickhelp',
        'parts',
        'modes_rectangle',
        'modes_diagonal',
        'modes_diagonal2',
        'staged',
    ]
    config = {
        default: kwargs[default]
        for default in default_list
        if default in kwargs
    }

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_detection', imgsetid=imgsetid, previous=gid, **config))


@register_route('/submit/viewpoint/', methods=['POST'])
def submit_viewpoint(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('viewpoint-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    aid = int(request.form['viewpoint-aid'])
    user_id = controller_inject.get_user().get('username', None)
    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        print('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid, ))
        aid = None  # Reset AID to prevent previous
    if method.lower() == 'make junk':
        ibs.set_annot_quality_texts([aid], [const.QUAL_JUNK])
        print('[web] (SET AS JUNK) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    if method.lower() == 'rotate left':
        ibs.update_annot_rotate_left_90([aid])
        print('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    if method.lower() == 'rotate right':
        ibs.update_annot_rotate_right_90([aid])
        print('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        viewpoint = int(request.form['viewpoint-value'])
        viewpoint_text = appf.VIEWPOINT_MAPPING.get(viewpoint, None)
        species_text = request.form['viewpoint-species']
        ibs.set_annot_viewpoints([aid], [viewpoint_text])
        # TODO ibs.set_annot_viewpoint_code([aid], [viewpoint_text])
        ibs.set_annot_species([aid], [species_text])
        print('[web] user_id: %s, aid: %d, viewpoint_text: %s' % (user_id, aid, viewpoint_text))
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_viewpoint', imgsetid=imgsetid, src_ag=src_ag,
                                dst_ag=dst_ag, previous=aid))


@register_route('/submit/viewpoint2/', methods=['POST'])
def submit_viewpoint2(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('viewpoint-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    aid = int(request.form['viewpoint-aid'])
    user_id = controller_inject.get_user().get('username', None)
    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        print('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid, ))
        aid = None  # Reset AID to prevent previous
    if method.lower() == 'make junk':
        ibs.set_annot_quality_texts([aid], [const.QUAL_JUNK])
        print('[web] (SET AS JUNK) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    if method.lower() == 'rotate left':
        ibs.update_annot_rotate_left_90([aid])
        print('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    if method.lower() == 'rotate right':
        ibs.update_annot_rotate_right_90([aid])
        print('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        if method.lower() == 'ignore':
            viewpoint = 'unknown'
        else:
            # Get metadata
            viewpoint1 = int(kwargs.get('ia-viewpoint-value-1', None))
            viewpoint2 = int(kwargs.get('ia-viewpoint-value-2', None))
            viewpoint3 = int(kwargs.get('ia-viewpoint-value-3', None))
            viewpoint_tup = (viewpoint1, viewpoint2, viewpoint3, )
            viewpoint = appf.convert_tuple_to_viewpoint(viewpoint_tup)
        ibs.set_annot_viewpoints([aid], [viewpoint])
        species_text = request.form['viewpoint-species']
        # TODO ibs.set_annot_viewpoint_code([aid], [viewpoint_text])
        ibs.set_annot_species([aid], [species_text])
        print('[web] user_id: %s, aid: %d, viewpoint_text: %s' % (user_id, aid, viewpoint))
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_viewpoint2', imgsetid=imgsetid, src_ag=src_ag,
                                dst_ag=dst_ag, previous=aid))


@register_route('/submit/viewpoint3/', methods=['POST'])
def submit_viewpoint3(**kwargs):
    with ut.Timer('[submit_viewpoint3]'):
        ibs = current_app.ibs
        method = request.form.get('viewpoint-submit', '')
        imgsetid = request.args.get('imgsetid', '')
        imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

        src_ag = request.args.get('src_ag', '')
        src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
        dst_ag = request.args.get('dst_ag', '')
        dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

        aid = int(request.form['viewpoint-aid'])
        user_id = controller_inject.get_user().get('username', None)

        if method.lower() == 'delete':
            ibs.delete_annots(aid)
            print('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid, ))
            aid = None  # Reset AID to prevent previous
        if method.lower() == 'make junk':
            ibs.set_annot_quality_texts([aid], [const.QUAL_JUNK])
            print('[web] (SET AS JUNK) user_id: %s, aid: %d' % (user_id, aid, ))
            redirection = request.referrer
            if 'aid' not in redirection:
                # Prevent multiple clears
                if '?' in redirection:
                    redirection = '%s&aid=%d' % (redirection, aid, )
                else:
                    redirection = '%s?aid=%d' % (redirection, aid, )
            return redirect(redirection)
        if method.lower() == 'rotate left':
            ibs.update_annot_rotate_left_90([aid])
            print('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid, ))
            redirection = request.referrer
            if 'aid' not in redirection:
                # Prevent multiple clears
                if '?' in redirection:
                    redirection = '%s&aid=%d' % (redirection, aid, )
                else:
                    redirection = '%s?aid=%d' % (redirection, aid, )
            return redirect(redirection)
        if method.lower() == 'rotate right':
            ibs.update_annot_rotate_right_90([aid])
            print('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid, ))
            redirection = request.referrer
            if 'aid' not in redirection:
                # Prevent multiple clears
                if '?' in redirection:
                    redirection = '%s&aid=%d' % (redirection, aid, )
                else:
                    redirection = '%s?aid=%d' % (redirection, aid, )
            return redirect(redirection)
        else:
            if src_ag is not None and dst_ag is not None:
                appf.movegroup_aid(ibs, aid, src_ag, dst_ag)

            if method.lower() == 'ignore':
                ibs.set_annot_viewpoints([aid], ['ignore'], only_allow_known=False, _code_update=False)
                viewpoint = 'ignore'
            else:
                # Get metadata
                viewpoint_str = kwargs.get('viewpoint-text-code', '')
                if not isinstance(viewpoint_str, six.string_types) or len(viewpoint_str) == 0:
                    viewpoint_str = None

                if viewpoint_str is None:
                    viewpoint = const.VIEW.UNKNOWN
                else:
                    viewpoint = getattr(const.VIEW, viewpoint_str, const.VIEW.UNKNOWN)
                ibs.set_annot_viewpoint_int([aid], [viewpoint])

            species_text = request.form['viewpoint-species']
            ibs.set_annot_species([aid], [species_text])
            ibs.set_annot_reviewed([aid], [1])
            print('[web] user_id: %s, aid: %d, viewpoint: %s' % (user_id, aid, viewpoint))

        # Return HTML
        refer = request.args.get('refer', '')
        if len(refer) > 0:
            retval = redirect(appf.decode_refer_url(refer))
        else:
            retval = redirect(url_for('turk_viewpoint3', imgsetid=imgsetid, src_ag=src_ag,
                                      dst_ag=dst_ag, previous=aid))

    return retval


@register_route('/submit/annotation/', methods=['POST'])
def submit_annotation(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('ia-annotation-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    aid = int(request.form['ia-annotation-aid'])
    user_id = controller_inject.get_user().get('username', None)
    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        print('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid, ))
        aid = None  # Reset AID to prevent previous
    elif method.lower() == 'make junk':
        ibs.set_annot_quality_texts([aid], [const.QUAL_JUNK])
        print('[web] (SET AS JUNK) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    elif method.lower() == u'rotate left':
        ibs.update_annot_rotate_left_90([aid])
        print('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    elif method.lower() == u'rotate right':
        ibs.update_annot_rotate_right_90([aid])
        print('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        try:
            viewpoint = int(request.form['ia-annotation-viewpoint-value'])
        except ValueError:
            viewpoint = int(float(request.form['ia-annotation-viewpoint-value']))
        viewpoint_text = appf.VIEWPOINT_MAPPING.get(viewpoint, None)
        species_text = request.form['ia-annotation-species']
        try:
            quality = int(request.form['ia-quality-value'])
        except ValueError:
            quality = int(float(request.form['ia-quality-value']))
        if quality in [-1, None]:
            quality = None
        elif quality == 0:
            quality = 2
        elif quality == 1:
            quality = 4
        else:
            raise ValueError('quality must be -1, 0 or 1')
        ibs.set_annot_viewpoints([aid], [viewpoint_text])
        # TODO ibs.set_annot_viewpoint_code([aid], [viewpoint_text])
        ibs.set_annot_species([aid], [species_text])
        ibs.set_annot_qualities([aid], [quality])
        multiple = 1 if 'ia-multiple-value' in request.form else 0
        ibs.set_annot_multiple([aid], [multiple])
        ibs.set_annot_reviewed([aid], [1])
        print('[web] user_id: %s, aid: %d, viewpoint: %r, quality: %r, multiple: %r' % (user_id, aid, viewpoint_text, quality, multiple))
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_annotation', imgsetid=imgsetid, src_ag=src_ag,
                                dst_ag=dst_ag, previous=aid))


@register_route('/submit/annotation/grid/', methods=['POST'])
def submit_annotation_grid(samples=200, species='zebra_grevys', version=1, **kwargs):
    ibs = current_app.ibs

    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    assert version in [1, 2, 3]

    aid_list = kwargs['annotation-grid-aids']
    highlight_list = kwargs['annotation-grid-highlighted']
    assert len(aid_list) == len(highlight_list)

    metadata_list = ibs.get_annot_metadata(aid_list)
    metadata_list_ = []
    for metadata, highlight in zip(metadata_list, highlight_list):
        if 'turk' not in metadata:
            metadata['turk'] = {}

        if version == 1:
            value = highlight
        elif version == 2:
            value = not highlight
        elif version == 3:
            value = highlight

        metadata['turk']['grid'] = value
        metadata_list_.append(metadata)

    ibs.set_annot_metadata(aid_list, metadata_list_)

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_annotation_grid',
                                imgsetid=imgsetid,
                                samples=samples, species=species,
                                version=version))


@register_route('/submit/splits/', methods=['POST'])
def submit_splits(**kwargs):
    ibs = current_app.ibs

    aid_list = kwargs['annotation-splits-aids']
    highlight_list = kwargs['annotation-splits-highlighted']
    assert len(aid_list) == len(highlight_list)

    ut.embed()

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_splits', aid=None))


@register_route('/submit/species/', methods=['POST'])
def submit_species(**kwargs):
    ibs = current_app.ibs

    method = request.form.get('ia-species-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    previous_species_rowids = request.form.get('ia-species-rowids', None)
    print('Using previous_species_rowids = %r' % (previous_species_rowids, ))

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    aid = int(request.form['ia-species-aid'])
    user_id = controller_inject.get_user().get('username', None)
    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        print('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid, ))
        aid = None  # Reset AID to prevent previous
    elif method.lower() == u'skip':
        print('[web] (SKIP) user_id: %s' % (user_id, ))
        return redirect(url_for('turk_species', imgsetid=imgsetid, src_ag=src_ag,
                                dst_ag=dst_ag, previous=aid,
                                previous_species_rowids=previous_species_rowids))
    elif method.lower() in u'refresh':
        print('[web] (REFRESH) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        if '?' in redirection:
            redirection = '%s&refresh=true' % (redirection, )
        else:
            redirection = '%s?refresh=true' % (redirection, )
        return redirect(redirection)
    elif method.lower() == u'rotate left':
        ibs.update_annot_rotate_left_90([aid])
        print('[web] (ROTATED LEFT) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    elif method.lower() == u'rotate right':
        ibs.update_annot_rotate_right_90([aid])
        print('[web] (ROTATED RIGHT) user_id: %s, aid: %d' % (user_id, aid, ))
        redirection = request.referrer
        if 'aid' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&aid=%d' % (redirection, aid, )
            else:
                redirection = '%s?aid=%d' % (redirection, aid, )
        return redirect(redirection)
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        # species_text = request.form['ia-species-species']
        species_text = kwargs.get('ia-species-value', '')
        if len(species_text) == 0:
            species_text = const.UNKNOWN
        ibs.set_annot_species([aid], [species_text])
        ibs.set_annot_reviewed([aid], [1])

        metadata_dict = ibs.get_annot_metadata(aid)
        if 'turk' not in metadata_dict:
            metadata_dict['turk'] = {}
        metadata_dict['turk']['species'] = user_id
        ibs.set_annot_metadata([aid], [metadata_dict])

        print('[web] user_id: %s, aid: %d, species: %r (%r)'  % (user_id, aid, species_text, metadata_dict, ))
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_species', imgsetid=imgsetid, src_ag=src_ag,
                                dst_ag=dst_ag, previous=aid,
                                previous_species_rowids=previous_species_rowids))


@register_route('/submit/quality/', methods=['POST'])
def submit_quality(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('quality-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)
    aid = int(request.form['quality-aid'])
    user_id = controller_inject.get_user().get('username', None)

    src_ag = request.args.get('src_ag', '')
    src_ag = None if src_ag == 'None' or src_ag == '' else int(src_ag)
    dst_ag = request.args.get('dst_ag', '')
    dst_ag = None if dst_ag == 'None' or dst_ag == '' else int(dst_ag)

    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        print('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid, ))
        aid = None  # Reset AID to prevent previous
    else:
        if src_ag is not None and dst_ag is not None:
            appf.movegroup_aid(ibs, aid, src_ag, dst_ag)
        quality = int(request.form['quality-value'])
        ibs.set_annot_qualities([aid], [quality])
        print('[web] user_id: %s, aid: %d, quality: %d' % (user_id, aid, quality))
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_quality', imgsetid=imgsetid, src_ag=src_ag,
                                dst_ag=dst_ag, previous=aid))


@register_route('/submit/demographics/', methods=['POST'])
def submit_demographics(species='zebra_grevys', **kwargs):
    ibs = current_app.ibs

    DAN_SPECIAL_WRITE_AGE_TO_ALL_ANOTATIONS = True

    method = request.form.get('demographics-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)
    aid = int(request.form['demographics-aid'])
    user_id = controller_inject.get_user().get('username', None)

    if method.lower() == 'delete':
        ibs.delete_annots(aid)
        print('[web] (DELETED) user_id: %s, aid: %d' % (user_id, aid, ))
        aid = None  # Reset AID to prevent previous
    else:
        sex = int(request.form['demographics-sex-value'])
        age = int(request.form['demographics-age-value'])
        age_min = None
        age_max = None
        # Sex
        if sex >= 2:
            sex -= 2
        else:
            sex = -1

        if age == 1:
            age_min = None
            age_max = None
        elif age == 2:
            age_min = None
            age_max = 2
        elif age == 3:
            age_min = 3
            age_max = 5
        elif age == 4:
            age_min = 6
            age_max = 11
        elif age == 5:
            age_min = 12
            age_max = 23
        elif age == 6:
            age_min = 24
            age_max = 35
        elif age == 7:
            age_min = 36
            age_max = None

        ibs.set_annot_sex([aid], [sex])
        nid = ibs.get_annot_name_rowids(aid)
        if nid is not None and DAN_SPECIAL_WRITE_AGE_TO_ALL_ANOTATIONS:
            aid_list = ibs.get_name_aids(nid)
        else:
            aid_list = [aid]

        ibs.set_annot_age_months_est_min(aid_list, [age_min] * len(aid_list))
        ibs.set_annot_age_months_est_max(aid_list, [age_max] * len(aid_list))
        print('[web] Updating %d demographics with user_id: %s\n\taid_list : %r\n\tsex: %r\n\tage_min: %r\n\tage_max: %r' % (len(aid_list), user_id, aid_list, sex, age_min, age_max,))
    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_demographics', imgsetid=imgsetid, previous=aid, species=species))


@register_route('/submit/identification/', methods=['POST'])
def submit_identification(**kwargs):
    from ibeis.web.apis_query import process_graph_match_html
    ibs = current_app.ibs

    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)
    aid1 = int(request.form['identification-aid1'])
    aid2 = int(request.form['identification-aid2'])
    replace_review_rowid = int(request.form.get('identification-replace-review-rowid', -1))

    # Process form data
    annot_uuid_1, annot_uuid_2, state, tag_list = process_graph_match_html(ibs)

    # Add state to staging database
    # FIXME:
    # photobomb and scenerymatch tags should be disjoint from match-state
    if state == 'matched':
        decision = const.EVIDENCE_DECISION.POSITIVE
    elif state == 'notmatched':
        decision = const.EVIDENCE_DECISION.NEGATIVE
    elif state == 'notcomparable':
        decision = const.EVIDENCE_DECISION.INCOMPARABLE
    elif state == 'photobomb':
        decision = const.EVIDENCE_DECISION.NEGATIVE
        tag_list = ['photobomb']
    elif state == 'scenerymatch':
        decision = const.EVIDENCE_DECISION.NEGATIVE
        tag_list = ['scenerymatch']
    else:
        raise ValueError()

    # Replace a previous decision
    if replace_review_rowid > 0:
        print('REPLACING OLD EVIDENCE_DECISION ID = %r' % (replace_review_rowid, ))
        ibs.delete_review([replace_review_rowid])

    # Add a new review row for the new decision (possibly replacing the old one)
    print('ADDING EVIDENCE_DECISION: %r %r %r %r' % (aid1, aid2, decision, tag_list, ))
    tags_list = None if tag_list is None else [tag_list]
    review_rowid = ibs.add_review([aid1], [aid2], [decision], tags_list=tags_list)
    review_rowid = review_rowid[0]
    previous = '%s;%s;%s' % (aid1, aid2, review_rowid, )

    # Notify any attached web QUERY_OBJECT
    try:
        state = const.EVIDENCE_DECISION.INT_TO_CODE[decision]
        feedback = (aid1, aid2, state, tags_list)
        print('Adding %r to QUERY_OBJECT_FEEDBACK_BUFFER' % (feedback, ))
        current_app.QUERY_OBJECT_FEEDBACK_BUFFER.append(feedback)
    except ValueError:
        pass

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_identification', imgsetid=imgsetid, previous=previous))


@register_route('/submit/identification/v2/', methods=['POST'])
def submit_identification_v2(graph_uuid, **kwargs):
    ibs = current_app.ibs

    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    # Process form data
    annot_uuid_1, annot_uuid_2 = ibs.process_graph_match_html_v2(graph_uuid, **kwargs)
    aid1 = ibs.get_annot_aids_from_uuid(annot_uuid_1)
    aid2 = ibs.get_annot_aids_from_uuid(annot_uuid_2)

    hogwild = kwargs.get('identification-hogwild', False)
    hogwild_species = kwargs.get('identification-hogwild-species', None)
    hogwild_species = None if hogwild_species == 'None' or hogwild_species == '' else hogwild_species
    print('Using hogwild: %r' % (hogwild, ))

    previous = '%s;%s;-1' % (aid1, aid2, )

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        base = url_for('turk_identification_graph')
        sep = '&' if '?' in base else '?'
        args = (base, sep, ut.to_json(graph_uuid), previous, hogwild, hogwild_species, )
        url = '%s%sgraph_uuid=%s&previous=%s&hogwild=%s&hogwild_species=%s' % args
        url = url.replace(': ', ':')
        return redirect(url)


@register_route('/submit/group_review/', methods=['POST'])
def group_review_submit(**kwargs):
    """
    CommandLine:
        python -m ibeis.web.app --exec-group_review_submit

    Example:
        >>> # UNSTABLE_DOCTEST
        >>> from ibeis.web.app import *  # NOQA
        >>> import ibeis
        >>> import ibeis.web
        >>> ibs = ibeis.opendb('testdb1')
        >>> aid_list = ibs.get_valid_aids()[::2]
        >>> ibs.start_web_annot_groupreview(aid_list)
    """
    ibs = current_app.ibs
    method = request.form.get('group-review-submit', '')
    if method.lower() == 'populate':
        redirection = request.referrer
        if 'prefill' not in redirection:
            # Prevent multiple clears
            if '?' in redirection:
                redirection = '%s&prefill=true' % (redirection, )
            else:
                redirection = '%s?prefill=true' % (redirection, )
        return redirect(redirection)
    aid_list = request.form.get('aid_list', '')
    if len(aid_list) > 0:
        aid_list = aid_list.replace('[', '')
        aid_list = aid_list.replace(']', '')
        aid_list = aid_list.strip().split(',')
        aid_list = [ int(aid_.strip()) for aid_ in aid_list ]
    else:
        aid_list = []
    src_ag, dst_ag = ibs.prepare_annotgroup_review(aid_list)
    valid_modes = ut.get_list_column(appf.VALID_TURK_MODES, 0)
    mode = request.form.get('group-review-mode', None)
    assert mode in valid_modes
    return redirect(url_for(mode, src_ag=src_ag, dst_ag=dst_ag))


@register_route('/submit/contour/', methods=['POST'])
def submit_contour(**kwargs):
    ibs = current_app.ibs
    method = request.form.get('contour-submit', '')
    imgsetid = request.args.get('imgsetid', '')
    imgsetid = None if imgsetid == 'None' or imgsetid == '' else int(imgsetid)

    part_rowid = int(request.form['contour-part-rowid'])

    if method.lower() == 'accept':
        data_dict = ut.from_json(request.form['ia-contour-data'])

        if data_dict is None:
            data_dict = {}

        contour_dict = ibs.get_part_contour(part_rowid)
        contour_dict['contour'] = data_dict
        ibs.set_part_contour([part_rowid], [contour_dict])

        segment = data_dict.get('segment', [])
        num_contours = 1
        num_points = len(segment)
        ibs.set_part_reviewed([part_rowid], [1])

        print('[web] part_rowid: %d, contours: %d, points: %d' % (part_rowid, num_contours, num_points, ))

    default_list = [
        'temp'
    ]
    config = {
        default: kwargs[default]
        for default in default_list
        if default in kwargs
    }

    # Return HTML
    refer = request.args.get('refer', '')
    if len(refer) > 0:
        return redirect(appf.decode_refer_url(refer))
    else:
        return redirect(url_for('turk_contour', imgsetid=imgsetid, previous=part_rowid, **config))


if __name__ == '__main__':
    """
    CommandLine:
        python -m ibeis.web.app
        python -m ibeis.web.app --allexamples
        python -m ibeis.web.app --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



"""
TODO: These tests are good and important to run.
Ensure they are run via run_tests even though they are not doctests.

Consider moving to pytest and using xdoctest (because regular doctest does not
accept the syntax of IBEIS doctests)
"""


def test_neg_metagraph_simple_add_remove():
    """
    Test that the negative metagraph tracks the number of negative edges
    between PCCs through non-label-changing operations
    """
    from ibeis.algo.graph import demo
    from ibeis.algo.graph.state import (POSTV, NEGTV, INCMP, UNREV, UNKWN)  # NOQA

    # Create a graph with 5-sized CCs, with 3-pos-redun, and no negative edges
    infr = demo.demodata_infr(num_pccs=2, pcc_size=5, pos_redun=3,
                              ignore_pair=True, infer=True)
    cc_a, cc_b, cc_c, cc_d = infr.positive_components()
    a1, a2, a3, a4, a5 = cc_a
    b1, b2, b3, b4, b5 = cc_b

    nmg = infr.neg_metagraph

    # Check there are 4 meta-nodes and no edges
    assert nmg.number_of_edges() == 0
    assert nmg.number_of_nodes() == 4

    # Should add 1 edge to the negative metagraph
    u, v = a1, b1
    infr.add_feedback((u, v), NEGTV)
    nid1, nid2 = infr.node_labels(u, v)
    assert nmg.edges[nid1, nid2]['weight'] == 1
    assert nmg.number_of_edges() == 1
    assert nmg.number_of_nodes() == 4

    # Adding a second time should do nothing
    edge = a1, b1
    infr.add_feedback(edge, NEGTV)
    name_edge = infr.node_labels(*edge)
    assert nmg.edges[name_edge]['weight'] == 1
    assert nmg.number_of_edges() == 1
    assert nmg.number_of_nodes() == 4

    # But adding a second between different nodes will increase the weight
    edge = a1, b2
    infr.add_feedback(edge, NEGTV)
    name_edge = infr.node_labels(*edge)
    assert nmg.edges[name_edge]['weight'] == 2
    assert nmg.number_of_edges() == 1
    assert nmg.number_of_nodes() == 4

    infr.add_feedback((u, v), NEGTV)
    assert nmg.edges[name_edge]['weight'] == 2

    # Removing or relabeling the edge will decrease the weight
    infr.add_feedback((a1, b2), INCMP)
    assert nmg.edges[name_edge]['weight'] == 1

    # And removing all will remove the negative edge
    infr.add_feedback((a1, b1), INCMP)
    assert not nmg.has_edge(*name_edge)

    infr.assert_neg_metagraph()


def test_neg_metagraph_merge():
    """
    Test that the negative metagraph tracks the number of negative edges
    between PCCs through label-changing merge operations
    """
    from ibeis.algo.graph import demo
    from ibeis.algo.graph.state import (POSTV, NEGTV, INCMP, UNREV, UNKWN)  # NOQA
    # Create a graph with 4 CCs, with 3-pos-redun, and no negative edges
    infr = demo.demodata_infr(num_pccs=4, pcc_size=5, pos_redun=3,
                              ignore_pair=True, infer=True)
    cc_a, cc_b, cc_c, cc_d = infr.positive_components()
    a1, a2, a3, a4, a5 = cc_a
    b1, b2, b3, b4, b5 = cc_b
    c1, c2, c3, c4, c5 = cc_c
    d1, d2, d3, d4, d5 = cc_d

    nmg = infr.neg_metagraph

    # Add three negative edges between a and b
    # one between (a, c), (b, d), (a, d), and (c, d)
    A, B, C, D = infr.node_labels(a1, b1, c1, d1)

    infr.add_feedback((a1, b1), NEGTV)
    infr.add_feedback((a2, b2), NEGTV)
    infr.add_feedback((a3, b3), NEGTV)
    infr.add_feedback((a4, c4), NEGTV)
    infr.add_feedback((b4, d4), NEGTV)
    infr.add_feedback((c1, d1), NEGTV)
    infr.add_feedback((a4, d4), NEGTV)

    assert nmg.edges[(A, B)]['weight'] == 3
    assert nmg.edges[(A, C)]['weight'] == 1
    assert (B, C) not in nmg.edges
    assert nmg.edges[(A, D)]['weight'] == 1
    assert nmg.edges[(B, D)]['weight'] == 1
    assert nmg.number_of_edges() == 5
    assert nmg.number_of_nodes() == 4

    # Now merge A and B
    infr.add_feedback((a1, b1), POSTV)
    AB = infr.node_label(a1)

    # The old meta-nodes should not be combined into AB
    assert infr.node_label(b1) == AB
    assert A != B
    assert A == AB or A not in nmg.nodes
    assert B == AB or B not in nmg.nodes

    # Should have combined weights from (A, D) and (B, D)
    # And (A, C) should be brought over as-is
    assert nmg.edges[(AB, D)]['weight'] == 2
    assert nmg.edges[(AB, C)]['weight'] == 1

    # should not have a self-loop weight weight 2
    # (it decreased because we changed a previously neg edge to pos)
    assert nmg.edges[(AB, AB)]['weight'] == 2
    assert len(list(nmg.selfloop_edges())) == 1

    # nothing should change between C and D
    assert nmg.edges[(C, D)]['weight'] == 1

    # Should decrease number of nodes and edges
    assert nmg.number_of_nodes() == 3
    assert nmg.number_of_edges() == 4

    infr.assert_neg_metagraph()

    # Additional merge
    infr.add_feedback((c2, d2), POSTV)
    CD = infr.node_label(c1)
    infr.assert_neg_metagraph()

    assert nmg.number_of_nodes() == 2
    assert nmg.number_of_edges() == 3

    assert nmg.edges[(CD, CD)]['weight'] == 1
    assert nmg.edges[(AB, CD)]['weight'] == 3
    assert nmg.edges[(AB, AB)]['weight'] == 2

    # Yet another merge
    infr.add_feedback((a1, c1), POSTV)
    ABCD = infr.node_label(c1)
    assert nmg.number_of_nodes() == 1
    assert nmg.number_of_edges() == 1
    nmg.edges[(ABCD, ABCD)]['weight'] = 6
    infr.assert_neg_metagraph()


def test_neg_metagraph_split_neg():
    """
    Test that the negative metagraph tracks the number of negative edges
    between PCCs through label-changing split operations
    """
    from ibeis.algo.graph import demo
    from ibeis.algo.graph.state import (POSTV, NEGTV, INCMP, UNREV, UNKWN)  # NOQA
    # Create a graph with 4 CCs, with 3-pos-redun, and no negative edges
    infr = demo.demodata_infr(num_pccs=4, pcc_size=5, pos_redun=3,
                              ignore_pair=True, infer=True)
    nmg = infr.neg_metagraph
    assert nmg.number_of_nodes() != infr.neg_graph.number_of_nodes()
    assert nmg.number_of_edges() == 0
    # remove all positive edges
    for edge in list(infr.pos_graph.edges()):
        infr.add_feedback(edge, NEGTV)
    # metagraph should not be isomorphic to infr.neg_graph
    assert nmg.number_of_nodes() == infr.neg_graph.number_of_nodes()
    assert nmg.number_of_edges() > 0
    assert nmg.number_of_edges() == infr.neg_graph.number_of_edges()
    infr.assert_neg_metagraph()


def test_neg_metagraph_split_incomp():
    from ibeis.algo.graph import demo
    from ibeis.algo.graph.state import (POSTV, NEGTV, INCMP, UNREV, UNKWN)  # NOQA
    infr = demo.demodata_infr(num_pccs=4, pcc_size=5, pos_redun=3,
                              ignore_pair=True, infer=True)
    nmg = infr.neg_metagraph
    assert nmg.number_of_nodes() < infr.neg_graph.number_of_nodes()
    assert nmg.number_of_edges() == 0
    # remove all positive edges
    for edge in list(infr.pos_graph.edges()):
        infr.add_feedback(edge, INCMP)
    # metagraph should not be isomorphic to infr.neg_graph
    assert nmg.number_of_nodes() == infr.neg_graph.number_of_nodes()
    assert nmg.number_of_edges() == 0
    infr.assert_neg_metagraph()


def test_neg_metagraph_split_and_merge():
    """
    Test that the negative metagraph tracks the number of negative edges
    between PCCs through label-changing split and merge operations
    """
    from ibeis.algo.graph import demo
    from ibeis.algo.graph.state import (POSTV, NEGTV, INCMP, UNREV, UNKWN)  # NOQA
    # Create a graph with 4 CCs, with 3-pos-redun, and no negative edges
    infr = demo.demodata_infr(num_pccs=4, pcc_size=5, pos_redun=3,
                              ignore_pair=True, infer=True)
    cc_a, cc_b, cc_c, cc_d = infr.positive_components()
    a1, a2, a3, a4, a5 = cc_a
    b1, b2, b3, b4, b5 = cc_b
    c1, c2, c3, c4, c5 = cc_c
    d1, d2, d3, d4, d5 = cc_d

    nmg = infr.neg_metagraph

    # Add three negative edges between a and b
    # one between (a, c), (b, d), (a, d), and (c, d)
    A, B, C, D = infr.node_labels(a1, b1, c1, d1)

    infr.add_feedback((a1, b1), NEGTV)
    infr.add_feedback((a2, b2), NEGTV)
    infr.add_feedback((a3, b3), NEGTV)
    infr.add_feedback((a4, c4), NEGTV)
    infr.add_feedback((b4, d4), NEGTV)
    infr.add_feedback((c1, d1), NEGTV)
    infr.add_feedback((a4, d4), NEGTV)

    assert nmg.edges[(A, B)]['weight'] == 3
    assert nmg.edges[(A, C)]['weight'] == 1
    assert (B, C) not in nmg.edges
    assert nmg.edges[(A, D)]['weight'] == 1
    assert nmg.edges[(B, D)]['weight'] == 1
    assert nmg.number_of_edges() == 5
    assert nmg.number_of_nodes() == 4

    # merge A and B
    infr.add_feedback((a1, b1), POSTV)
    assert nmg.number_of_edges() == 4
    assert nmg.number_of_nodes() == 3
    AB = infr.node_label(a1)
    assert nmg.edges[(AB, AB)]['weight'] == 2

    # split A and B
    # the number of nodes should increase, but the edges should stay the
    # same because we added an incmp edge
    infr.add_feedback((a1, b1), INCMP)
    assert nmg.number_of_edges() == 5
    assert nmg.number_of_nodes() == 4
    assert nmg.edges[(A, B)]['weight'] == 2
    infr.assert_neg_metagraph()

    # remove all positive edges
    for edge in list(infr.pos_graph.edges()):
        infr.add_feedback(edge, INCMP)

    # metagraph should not be isomorphic to infr.neg_graph
    assert nmg.number_of_nodes() == infr.neg_graph.number_of_nodes()
    assert nmg.number_of_edges() == infr.neg_graph.number_of_edges()
    infr.assert_neg_metagraph()
